"""
This module defines the ACI_EXECUTE_FUNCTION meta function, which is used by LLM to submit
execution requests for indexed functions on aipolabs ACI backend.

This module includes the schema definition for the function and a Pydantic model for
validating the execution parameters.
"""

import aci.meta_functions._aci_get_function_definition as ACIGetFunctionDefinition

NAME = "ACI_EXECUTE_FUNCTION"
SCHEMA = {
    "type": "function",
    "function": {
        "name": NAME,
        "description": "Execute a specific retrieved function. Provide the executable function name, and the "
        "required function parameters for that function based on function definition retrieved.",
        "parameters": {
            "type": "object",
            "properties": {
                "function_name": {
                    "type": "string",
                    "description": f"The name of the function to execute, which is retrieved from the "
                    f"{ACIGetFunctionDefinition.NAME} function.",
                },
                "function_arguments": {
                    "type": "object",
                    "description": "A dictionary containing key-value pairs of input parameters required by the "
                    "specified function. The parameter names and types must match those defined in "
                    "the function definition previously retrieved. If the function requires no "
                    "parameters, provide an empty object.",
                    "additionalProperties": True,
                },
            },
            "required": ["function_name", "function_arguments"],
            "additionalProperties": False,
        },
    },
}


def wrap_function_arguments_if_not_present(obj: dict) -> dict:
    if "function_arguments" not in obj:
        # Create a copy of the input dict
        processed_obj = obj.copy()
        if "function_name" not in processed_obj:
            raise ValueError("function_name is required")
        # Extract function_name
        function_name = processed_obj.pop("function_name")
        # Create new dict with correct structure
        processed_obj = {
            "function_name": function_name,
            "function_arguments": processed_obj,  # All remaining fields go here
        }
        return processed_obj
    return obj



import httpx
import pytest
import respx

from aci import ACI
from aci._constants import DEFAULT_MAX_RETRIES
from aci._exceptions import (
    AuthenticationError,
    NotFoundError,
    PermissionError,
    RateLimitError,
    ServerError,
    UnknownError,
    ValidationError,
)
from aci.types.functions import FunctionDefinitionFormat

from .utils import MOCK_BASE_URL

MOCK_LINKED_ACCOUNT_OWNER_ID = "123"
MOCK_FUNCTION_NAME = "TEST_FUNCTION"
MOCK_FUNCTION_ARGUMENTS = {"param1": "value1", "param2": "value2"}


@respx.mock
@pytest.mark.parametrize(
    "search_params",
    [
        {},
        {
            "app_names": ["TEST"],
            "intent": "test",
            "allowed_apps_only": True,
            "format": FunctionDefinitionFormat.OPENAI,
            "limit": 10,
            "offset": 0,
        },
    ],
)
def test_search_functions_success(client: ACI, search_params: dict) -> None:
    mock_response = [{"name": "string", "description": "string"}]

    route = respx.get(f"{MOCK_BASE_URL}functions/search").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    functions = client.functions.search(**search_params)
    assert functions == mock_response
    assert route.call_count == 1, "should not retry"


@respx.mock
@pytest.mark.parametrize(
    "format, mock_response",
    [
        (
            FunctionDefinitionFormat.OPENAI,
            {
                "type": "function",
                "function": {
                    "name": "function_name",
                    "description": "function_description",
                    "parameters": {},
                },
            },
        ),
        (
            FunctionDefinitionFormat.ANTHROPIC,
            {
                "name": "function_name",
                "description": "function_description",
                "input_schema": {},
            },
        ),
    ],
)
def test_get_function_definition_success(
    client: ACI, format: FunctionDefinitionFormat, mock_response: dict
) -> None:
    route = respx.get(
        f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/definition",
        params={"format": format.value},
    ).mock(return_value=httpx.Response(200, json=mock_response))

    response = client.functions.get_definition(MOCK_FUNCTION_NAME, format)
    assert response == mock_response
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_get_function_definition_unauthorized(client: ACI) -> None:
    route = respx.get(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/definition").mock(
        return_value=httpx.Response(401, json={"message": "Unauthorized"})
    )

    with pytest.raises(AuthenticationError) as exc_info:
        client.functions.get_definition(MOCK_FUNCTION_NAME)

    assert "Unauthorized" in str(exc_info.value)
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_get_function_definition_forbidden(client: ACI) -> None:
    route = respx.get(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/definition").mock(
        return_value=httpx.Response(403, json={"message": "Forbidden"})
    )

    with pytest.raises(PermissionError) as exc_info:
        client.functions.get_definition(MOCK_FUNCTION_NAME)

    assert "Forbidden" in str(exc_info.value)
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_get_function_definition_not_found(client: ACI) -> None:
    route = respx.get(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/definition").mock(
        return_value=httpx.Response(404, json={"message": "Function not found"})
    )

    with pytest.raises(NotFoundError) as exc_info:
        client.functions.get_definition(MOCK_FUNCTION_NAME)

    assert "Function not found" in str(exc_info.value)
    assert route.call_count == 1, "should not retry"


@respx.mock
@pytest.mark.parametrize(
    "function_arguments",
    [
        {},
        MOCK_FUNCTION_ARGUMENTS,
    ],
)
def test_execute_function_success(client: ACI, function_arguments: dict) -> None:
    mock_response = {"success": True, "data": "string"}
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    response = client.functions.execute(
        MOCK_FUNCTION_NAME, function_arguments, MOCK_LINKED_ACCOUNT_OWNER_ID
    )
    assert response.model_dump(exclude_none=True) == mock_response
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_execute_function_bad_request(client: ACI) -> None:
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        return_value=httpx.Response(400, json={"message": "Bad request"})
    )

    with pytest.raises(ValidationError) as exc_info:
        client.functions.execute(
            MOCK_FUNCTION_NAME, MOCK_FUNCTION_ARGUMENTS, MOCK_LINKED_ACCOUNT_OWNER_ID
        )

    assert "Bad request" in str(exc_info.value)
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_execute_function_rate_limit_exceeded(client: ACI) -> None:
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        return_value=httpx.Response(429, json={"message": "Rate limit exceeded"})
    )

    with pytest.raises(RateLimitError) as exc_info:
        client.functions.execute(
            MOCK_FUNCTION_NAME, MOCK_FUNCTION_ARGUMENTS, MOCK_LINKED_ACCOUNT_OWNER_ID
        )

    assert "Rate limit exceeded" in str(exc_info.value)
    assert route.call_count == DEFAULT_MAX_RETRIES, "should retry"


@respx.mock
def test_execute_function_server_error(client: ACI) -> None:
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        return_value=httpx.Response(500, json={"message": "Internal server error"})
    )

    with pytest.raises(ServerError) as exc_info:
        client.functions.execute(
            MOCK_FUNCTION_NAME, MOCK_FUNCTION_ARGUMENTS, MOCK_LINKED_ACCOUNT_OWNER_ID
        )

    assert route.call_count == DEFAULT_MAX_RETRIES, "should retry"
    assert "Internal server error" in str(exc_info.value)


@respx.mock
def test_execute_function_unknown_error(client: ACI) -> None:
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        return_value=httpx.Response(418, json={"message": "I'm a teapot"})
    )

    with pytest.raises(UnknownError):
        client.functions.execute(
            MOCK_FUNCTION_NAME, MOCK_FUNCTION_ARGUMENTS, MOCK_LINKED_ACCOUNT_OWNER_ID
        )

    assert route.call_count == DEFAULT_MAX_RETRIES, "should retry"


@respx.mock
def test_execute_function_timeout_exception(client: ACI) -> None:
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        side_effect=httpx.TimeoutException("Request timed out")
    )

    with pytest.raises(httpx.TimeoutException) as exc_info:
        client.functions.execute(
            MOCK_FUNCTION_NAME, MOCK_FUNCTION_ARGUMENTS, MOCK_LINKED_ACCOUNT_OWNER_ID
        )

    assert "Request timed out" in str(exc_info.value)
    assert route.call_count == DEFAULT_MAX_RETRIES, "should retry"


@respx.mock
def test_execute_function_network_error(client: ACI) -> None:
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        side_effect=httpx.NetworkError("Network error")
    )

    with pytest.raises(httpx.NetworkError) as exc_info:
        client.functions.execute(
            MOCK_FUNCTION_NAME, MOCK_FUNCTION_ARGUMENTS, MOCK_LINKED_ACCOUNT_OWNER_ID
        )

    assert "Network error" in str(exc_info.value)
    assert route.call_count == DEFAULT_MAX_RETRIES, "should retry"


@respx.mock
def test_execute_function_retry_on_server_error(client: ACI) -> None:
    mock_success_response = {"success": True, "data": "string"}

    # Simulate two server errors followed by a successful response
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        side_effect=[
            httpx.Response(500, json={"message": "Internal server error"}),
            httpx.Response(500, json={"message": "Internal server error"}),
            httpx.Response(200, json=mock_success_response),
        ]
    )

    response = client.functions.execute(
        MOCK_FUNCTION_NAME, MOCK_FUNCTION_ARGUMENTS, MOCK_LINKED_ACCOUNT_OWNER_ID
    )
    assert route.call_count == 3, "should retry until success"
    assert response.model_dump(exclude_none=True) == mock_success_response


@respx.mock
def test_execute_function_retry_exhausted(client: ACI) -> None:
    route = respx.post(f"{MOCK_BASE_URL}functions/{MOCK_FUNCTION_NAME}/execute").mock(
        side_effect=[
            httpx.Response(500, json={"message": "Internal server error"}),
            httpx.Response(500, json={"message": "Internal server error"}),
            httpx.Response(500, json={"message": "Internal server error"}),
            httpx.Response(500, json={"message": "Internal server error"}),
            httpx.Response(500, json={"message": "Internal server error"}),
        ]
    )

    with pytest.raises(ServerError) as exc_info:
        client.functions.execute(
            MOCK_FUNCTION_NAME, MOCK_FUNCTION_ARGUMENTS, MOCK_LINKED_ACCOUNT_OWNER_ID
        )

    assert route.call_count == DEFAULT_MAX_RETRIES, "should retry"
    assert "Internal server error" in str(exc_info.value)



from typing import Generator

import pytest

from aci import ACI

from .utils import MOCK_API_KEY, MOCK_BASE_URL


@pytest.fixture(scope="session")
def client() -> Generator[ACI, None, None]:
    with ACI(api_key=MOCK_API_KEY, base_url=MOCK_BASE_URL) as client:
        yield client



import os

import pytest

from aci import ACI
from aci.types.enums import SecurityScheme

ACI_API_KEY = os.getenv("ACI_API_KEY")

"""
The repo doesn't use load_dotenv() so we need to set the environment variables manually.
You can use something like dotenv to load the environment variables from a .env file for a single run.
e.g.

```
$ npm install -g dotenv-cli
$ dotenv run pytest tests/e2e/test_app_configurations.py -s
```
"""


@pytest.mark.skipif(ACI_API_KEY is None, reason="ACI_API_KEY is not set")
def test_app_configurations() -> None:
    aci = ACI(api_key=ACI_API_KEY)
    # step 0: make sure no app configurations exist
    assert len(aci.app_configurations.list()) == 0

    # step 1: app configurations of different types of security schemes
    gmail_configuration = aci.app_configurations.create(
        app_name="GMAIL",
        security_scheme=SecurityScheme.OAUTH2,
    )
    assert gmail_configuration.app_name == "GMAIL"
    assert gmail_configuration.security_scheme == SecurityScheme.OAUTH2

    brave_search_configuration = aci.app_configurations.create(
        app_name="BRAVE_SEARCH",
        security_scheme=SecurityScheme.API_KEY,
    )
    assert brave_search_configuration.app_name == "BRAVE_SEARCH"
    assert brave_search_configuration.security_scheme == SecurityScheme.API_KEY

    asm_configuration = aci.app_configurations.create(
        app_name="AIPOLABS_SECRETS_MANAGER",
        security_scheme=SecurityScheme.NO_AUTH,
    )
    assert asm_configuration.app_name == "AIPOLABS_SECRETS_MANAGER"
    assert asm_configuration.security_scheme == SecurityScheme.NO_AUTH

    # step 2: get a specific app configuration
    retrieved_gmail_configuration = aci.app_configurations.get(app_name="GMAIL")
    assert retrieved_gmail_configuration.app_name == "GMAIL"
    assert retrieved_gmail_configuration.security_scheme == SecurityScheme.OAUTH2

    # step 3: list all app configurations
    app_configurations = aci.app_configurations.list()
    assert len(app_configurations) == 3
    assert all(
        app_configuration.app_name in ["GMAIL", "BRAVE_SEARCH", "AIPOLABS_SECRETS_MANAGER"]
        for app_configuration in app_configurations
    )
    assert all(
        app_configuration.security_scheme
        in [SecurityScheme.OAUTH2, SecurityScheme.API_KEY, SecurityScheme.NO_AUTH]
        for app_configuration in app_configurations
    )

    # step 4: list with app name filter
    app_configurations = aci.app_configurations.list(app_names=["GMAIL", "BRAVE_SEARCH"])
    assert len(app_configurations) == 2
    assert all(
        app_configuration.app_name in ["GMAIL", "BRAVE_SEARCH"]
        for app_configuration in app_configurations
    )

    # step 5: list with limit
    app_configurations = aci.app_configurations.list(limit=1)
    assert len(app_configurations) == 1
    assert app_configurations[0].app_name in [
        "GMAIL",
        "BRAVE_SEARCH",
        "AIPOLABS_SECRETS_MANAGER",
    ]

    # step 6: delete all app configurations
    for app_name in ["GMAIL", "BRAVE_SEARCH", "AIPOLABS_SECRETS_MANAGER"]:
        aci.app_configurations.delete(app_name=app_name)
    assert len(aci.app_configurations.list()) == 0



import httpx
import pytest
import respx

from aci import ACI

from .utils import MOCK_BASE_URL


@respx.mock
@pytest.mark.parametrize(
    "search_params",
    [
        {},  # Base case: No optional parameters provided.
        {  # All optional parameters provided.
            "intent": "test",
            "allowed_apps_only": True,
            "include_functions": True,
            "categories": ["utility", "education"],
            "limit": 10,
            "offset": 5,
        },
    ],
)
def test_search_apps_success(client: ACI, search_params: dict) -> None:
    mock_response = [
        {
            "name": "string",
            "description": "string",
            "functions": [{"name": "string", "description": "string"}],
        }
    ]

    route = respx.get(f"{MOCK_BASE_URL}apps/search").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    apps = client.apps.search(**search_params)
    assert [app.model_dump(exclude_none=True) for app in apps] == mock_response
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_get_app_success(client: ACI) -> None:
    app_name = "TEST_APP"
    mock_response = {
        "id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "name": "string",
        "display_name": "string",
        "provider": "string",
        "version": "string",
        "description": "string",
        "logo": "string",
        "categories": ["string"],
        "visibility": "public",
        "active": True,
        "security_schemes": ["no_auth"],
        "functions": [
            {
                "id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
                "app_name": "string",
                "name": "string",
                "description": "string",
                "tags": ["string"],
                "visibility": "public",
                "active": True,
                "protocol": "rest",
                "protocol_data": {},
                "parameters": {},
                "response": {},
            }
        ],
    }
    route = respx.get(f"{MOCK_BASE_URL}apps/{app_name}").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    app = client.apps.get(app_name)
    assert app.model_dump() == mock_response
    assert route.call_count == 1, "should not retry"



import httpx
import respx

from aci import ACI
from aci.meta_functions import (
    ACIExecuteFunction,
    ACIGetFunctionDefinition,
    ACISearchApps,
    ACISearchFunctions,
)

from .utils import MOCK_BASE_URL, MOCK_LINKED_ACCOUNT_OWNER_ID


@respx.mock
def test_handle_function_call_search_apps(client: ACI) -> None:
    mock_response = [{"name": "Test App", "description": "Test Description"}]

    route = respx.get(f"{MOCK_BASE_URL}apps/search").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    response = client.handle_function_call(
        ACISearchApps.NAME,
        {"intent": "search apps"},
        linked_account_owner_id=MOCK_LINKED_ACCOUNT_OWNER_ID,
    )
    assert response == mock_response
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_handle_function_call_search_functions(client: ACI) -> None:
    mock_response = [{"name": "Test Function", "description": "Test Description"}]

    route = respx.get(f"{MOCK_BASE_URL}functions/search").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    response = client.handle_function_call(
        ACISearchFunctions.NAME,
        {"app_names": ["TEST"], "intent": "search functions"},
        linked_account_owner_id=MOCK_LINKED_ACCOUNT_OWNER_ID,
    )
    assert response == mock_response
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_handle_function_call_get_function_definition(client: ACI) -> None:
    function_arguments = {"function_name": "TEST_FUNCTION"}
    mock_response = {"function": {"name": "Test Function"}}
    # note: the function name for mock route here should be function_name in the function_arguments
    route = respx.get(
        f"{MOCK_BASE_URL}functions/{function_arguments['function_name']}/definition"
    ).mock(return_value=httpx.Response(200, json=mock_response))

    response = client.handle_function_call(
        ACIGetFunctionDefinition.NAME,
        function_arguments,
        linked_account_owner_id=MOCK_LINKED_ACCOUNT_OWNER_ID,
    )
    assert isinstance(response, dict)
    assert response == mock_response
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_handle_function_call_meta_function_execution(client: ACI) -> None:
    function_arguments = {
        "function_name": "BRAVE_SEARCH__WEB_SEARCH",
        "function_arguments": {"param1": "value1"},
    }
    mock_response = {"success": True, "data": "string"}
    # note: the function name for mock route here should be function_name in the function_arguments
    route = respx.post(
        f"{MOCK_BASE_URL}functions/{function_arguments['function_name']}/execute"
    ).mock(return_value=httpx.Response(200, json=mock_response))

    response = client.handle_function_call(
        ACIExecuteFunction.NAME,
        function_arguments,
        linked_account_owner_id=MOCK_LINKED_ACCOUNT_OWNER_ID,
    )
    assert response == mock_response
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_handle_function_call_direct_indexed_function_execution(client: ACI) -> None:
    function_name = "BRAVE_SEARCH__WEB_SEARCH"
    function_arguments = {"query": "test"}
    mock_response = {
        "success": True,
        "data": {"results": [{"title": "Test Result"}], "metadata": None},
    }

    route = respx.post(f"{MOCK_BASE_URL}functions/{function_name}/execute").mock(
        return_value=httpx.Response(200, json=mock_response)
    )
    response = client.handle_function_call(
        function_name,
        function_arguments,
        linked_account_owner_id=MOCK_LINKED_ACCOUNT_OWNER_ID,
    )

    assert response == mock_response
    assert route.call_count == 1, "should not retry"



from enum import Enum


class SecurityScheme(str, Enum):
    """
    security scheme type for an app
    """

    NO_AUTH = "no_auth"
    API_KEY = "api_key"
    HTTP_BASIC = "http_basic"
    HTTP_BEARER = "http_bearer"
    OAUTH2 = "oauth2"


class Visibility(str, Enum):
    """Visibility of an app or function."""

    PUBLIC = "public"
    PRIVATE = "private"
